==> hlog-0.17.tar.gz: hlog-0.17/daemon/location.c <==
/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <unistd.h>
#include <gps.h>

#include <jeffpc/error.h>
#include <jeffpc/thread.h>
#include <jeffpc/time.h>

#include <hlog/maidenhead.h>

#include "announce.h"
#include "location.h"

#define HOSTNAME	"localhost"
#define PORT		"2947"

/* How long to wait between ANNOUNCE_LOC_CURRENT notifications */
#define UPDATE_TIMEOUT	(60ull * 1000 * 1000 * 1000) /* ns */

/* How long to wait between reconnect attempts */
#define RECONNECT_DELAY 60 /* s */

struct fix {
	int mode;
	double lat;
	double lon;
};

static struct gps_data_t gps;
static struct fix prevfix;
static bool connected;
static pthread_t location_thread;

static void print_notification(enum announce_msg_type type, int mode,
			       char *grid)
{
	switch (type) {
		case ANNOUNCE_LOC_CURRENT:
			return; /* don't print periodic updates */
		case ANNOUNCE_LOC_ACQ:
			cmn_err(CE_INFO, "Acquired %dD fix in %s",
				mode, grid);
			break;
		case ANNOUNCE_LOC_CHANGE:
			cmn_err(CE_INFO, "Moved to %dD fix in %s",
				mode, grid);
			break;
		case ANNOUNCE_LOC_LOST:
			cmn_err(CE_INFO, "Fix lost");
			break;
		case ANNOUNCE_LOC_DISCONNECT:
			cmn_err(CE_INFO, "Connection to gpsd lost");
			break;
		case ANNOUNCE_QSO_NEW:
		case ANNOUNCE_QSO_UPDATED:
			panic("Invalid type specified (%d)", type);
	}
}

static void notify(enum announce_msg_type type, struct fix *fix)
{
	static uint64_t last_update;
	char grid[MAIDENHEAD_MAX_STRING_LENGTH];

	switch (type) {
		case ANNOUNCE_LOC_CURRENT:
			if ((last_update + UPDATE_TIMEOUT) > gettime())
				return; /* not enough time elapsed */
			/* fallthru */
		case ANNOUNCE_LOC_ACQ:
		case ANNOUNCE_LOC_CHANGE:
			if (!fix->mode) {
				memset(grid, 0, sizeof(grid));
			} else {
				if (!maidenhead_encode(grid, sizeof(grid),
						       fix->lat, fix->lon)) {
					cmn_err(CE_WARN, "Failed to convert "
						"location notification lat/lon "
						"to grid");
					return;
				}
			}

			print_notification(type, fix->mode, grid);
			announce_notify_loc(type, fix->mode, grid);
			last_update = gettime();
			return;
		case ANNOUNCE_LOC_LOST:
		case ANNOUNCE_LOC_DISCONNECT:
			print_notification(type, 0, NULL);
			announce_notify_loc(type, 0, NULL);
			last_update = gettime();
			return;
		case ANNOUNCE_QSO_NEW:
		case ANNOUNCE_QSO_UPDATED:
			panic("invalid type specified (%d)", type);
	}

	panic("unknown type specified (%d)", type);
}

static void process(void)
{
	struct fix current;
	char pgrid[MAIDENHEAD_MAX_STRING_LENGTH];
	char cgrid[MAIDENHEAD_MAX_STRING_LENGTH];
	struct fix *p = &prevfix;
	struct fix *c = &current;

	if (!(gps.set & MODE_SET))
		return;

	current.mode = (gps.fix.mode <= 1) ? 0 : gps.fix.mode; /* 0, 2, 3 */
	current.lat = gps.fix.latitude;
	current.lon = gps.fix.longitude;

	ASSERT3U(!current.mode, ==, !(gps.set & LATLON_SET));

	/* mode changed, inform the client as needed */
	if (p->mode != c->mode) {
		/* none -> 2D/3D? */
		if (!p->mode && c->mode)
			notify(ANNOUNCE_LOC_ACQ, c);

		/* 2D/3D -> none? */
		if (p->mode && !c->mode)
			/* fix lost */
			notify(ANNOUNCE_LOC_LOST, NULL);

		/* 2D/3D -> 3D/2D? */
		if (p->mode && c->mode)
			notify(ANNOUNCE_LOC_CHANGE, c);

		goto out;
	}

	/*
	 * figure out if the grid changed
	 *
	 * The simplest way to do this is to generate the strings for the
	 * old and the new, and compare them.
	 */
	if (c->mode &&
	    maidenhead_encode(pgrid, sizeof(pgrid), p->lat, p->lon) &&
	    maidenhead_encode(cgrid, sizeof(cgrid), c->lat, c->lon) &&
	    strcmp(pgrid, cgrid)) {
		notify(ANNOUNCE_LOC_CHANGE, c);
		goto out;
	}

	/*
	 * periodically send out the location
	 *
	 * Even if:
	 *  - the grid didn't change
	 *  - we don't have a fix
	 *
	 * This way, we're not silent for long periods without changes to
	 * the fix.
	 */
	notify(ANNOUNCE_LOC_CURRENT, c);

out:
	*p = *c;
}

static bool __gps_connect(void)
{
	/*
	 * To avoid spamming the log with reconnection attempts, we only
	 * print the error messages only once until we get a connection.
	 * Then, if we lose the connection, we print the messages once.  And
	 * so on.
	 */
	static bool silence;

	if (connected)
		return true;

	if (gps_open(HOSTNAME, PORT, &gps)) {
		if (!silence)
			cmn_err(CE_WARN, "Connection to gpsd failed: %s",
				gps_errstr(errno));
		silence = true;
		return false;
	}

	if (gps_stream(&gps, WATCH_ENABLE | WATCH_JSON, NULL)) {
		cmn_err(CE_WARN, "Failed to enable gpsd watches: %s",
			gps_errstr(errno));

		if (gps_close(&gps))
			cmn_err(CE_ERROR, "Failed to close gpsd connection: %s",
				gps_errstr(errno));

		silence = true;
		return false;
	}

	cmn_err(CE_INFO, "Connected, acquiring location info");

	connected = true;
	silence = false;

	/* reset previous fix data */
	memset(&prevfix, 0, sizeof(prevfix));

	return true;
}

/*
 * returns:
 *   -1 on error
 *    0 if no data available
 *    1 if data availabl
 */
static int __gps_read(void)
{
	int ret;

	if (!gps_waiting(&gps, 500))
		return 0;

	errno = 0;
	ret = gps_read(&gps, NULL, 0);
	if (ret < 0) {
		connected = false;
		notify(ANNOUNCE_LOC_DISCONNECT, NULL);
		return -1; /* error */
	} else if (!ret) {
		return 0; /* no data available */
	} else {
		return 1; /* data available */
	}
}

static void *location_watcher(void *unused)
{
	for (;;) {
		/* connect as necessary */
		if (!__gps_connect()) {
			/* failed - sleep for a while and try again */
			sleep(RECONNECT_DELAY);
			continue;
		}

		switch (__gps_read()) {
			case -1:
			case 0:
				sleep(1); /* nothing to do */
				break;
			case 1:
				process();
				break;
		}
	}

	cmn_err(CE_INFO, "Thread exiting...");

	notify(ANNOUNCE_LOC_DISCONNECT, NULL);

	/* clean up libgps */
	if (gps_close(&gps))
		cmn_err(CE_ERROR, "Failed to close gpsd connection: %s",
			strerror(errno));

	return NULL;
}

int location_start(void)
{
	int ret;

	ret = xthr_create_named(&location_thread, "location",
				location_watcher, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to start location thread: %s",
			xstrerror(ret));

	return ret;
}

void location_stop(void)
{
	int ret;

	ret = xthr_join(location_thread, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to join location thread: %s",
			xstrerror(ret));
}

==> hlog-0.20-rc1.tar.gz: hlog-0.20-rc1/daemon/location.c <==
/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <unistd.h>
#include <gps.h>

#include <jeffpc/error.h>
#include <jeffpc/thread.h>
#include <jeffpc/time.h>

#include <hlog/maidenhead.h>
#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-pub.h>

#include "location.h"

#define HOSTNAME	"localhost"
#define PORT		"2947"

/* How long to wait between ANNOUNCE_LOC_CURRENT notifications */
#define UPDATE_TIMEOUT	(60ull * 1000 * 1000 * 1000) /* ns */

/* How long to wait between reconnect attempts */
#define RECONNECT_DELAY 60 /* s */

struct fix {
	int mode;
	double lat;
	double lon;
};

static struct gps_data_t gps;
static struct fix prevfix;
static bool connected;
static pthread_t location_thread;

static void print_notification(enum announce_loc_msg_type type, int mode,
			       char *grid)
{
	switch (type) {
		case ANNOUNCE_LOC_CURRENT:
			return; /* don't print periodic updates */
		case ANNOUNCE_LOC_ACQ:
			cmn_err(CE_INFO, "Acquired %dD fix in %s",
				mode, grid);
			break;
		case ANNOUNCE_LOC_CHANGE:
			cmn_err(CE_INFO, "Moved to %dD fix in %s",
				mode, grid);
			break;
		case ANNOUNCE_LOC_LOST:
			cmn_err(CE_INFO, "Fix lost");
			break;
		case ANNOUNCE_LOC_DISCONNECT:
			cmn_err(CE_INFO, "Connection to gpsd lost");
			break;
	}
}

static void notify(enum announce_loc_msg_type type, struct fix *fix)
{
	static uint64_t last_update;
	char grid[MAIDENHEAD_MAX_STRING_LENGTH];

	switch (type) {
		case ANNOUNCE_LOC_CURRENT:
			if ((last_update + UPDATE_TIMEOUT) > gettime())
				return; /* not enough time elapsed */
			/* fallthru */
		case ANNOUNCE_LOC_ACQ:
		case ANNOUNCE_LOC_CHANGE:
			if (!fix->mode) {
				memset(grid, 0, sizeof(grid));
			} else {
				if (!maidenhead_encode(grid, sizeof(grid),
						       fix->lat, fix->lon)) {
					cmn_err(CE_WARN, "Failed to convert "
						"location notification lat/lon "
						"to grid");
					return;
				}
			}

			print_notification(type, fix->mode, grid);
			announce_publish_loc(type, fix->mode, grid);
			last_update = gettime();
			return;
		case ANNOUNCE_LOC_LOST:
		case ANNOUNCE_LOC_DISCONNECT:
			print_notification(type, 0, NULL);
			announce_publish_loc(type, 0, NULL);
			last_update = gettime();
			return;
	}

	panic("unknown type specified (%d)", type);
}

static void process(void)
{
	struct fix current;
	char pgrid[MAIDENHEAD_MAX_STRING_LENGTH];
	char cgrid[MAIDENHEAD_MAX_STRING_LENGTH];
	struct fix *p = &prevfix;
	struct fix *c = &current;

	if (!(gps.set & MODE_SET))
		return;

	current.mode = (gps.fix.mode <= 1) ? 0 : gps.fix.mode; /* 0, 2, 3 */
	current.lat = gps.fix.latitude;
	current.lon = gps.fix.longitude;

	ASSERT3U(!current.mode, ==, !(gps.set & LATLON_SET));

	/* mode changed, inform the client as needed */
	if (p->mode != c->mode) {
		/* none -> 2D/3D? */
		if (!p->mode && c->mode)
			notify(ANNOUNCE_LOC_ACQ, c);

		/* 2D/3D -> none? */
		if (p->mode && !c->mode)
			/* fix lost */
			notify(ANNOUNCE_LOC_LOST, NULL);

		/* 2D/3D -> 3D/2D? */
		if (p->mode && c->mode)
			notify(ANNOUNCE_LOC_CHANGE, c);

		goto out;
	}

	/*
	 * figure out if the grid changed
	 *
	 * The simplest way to do this is to generate the strings for the
	 * old and the new, and compare them.
	 */
	if (c->mode &&
	    maidenhead_encode(pgrid, sizeof(pgrid), p->lat, p->lon) &&
	    maidenhead_encode(cgrid, sizeof(cgrid), c->lat, c->lon) &&
	    strcmp(pgrid, cgrid)) {
		notify(ANNOUNCE_LOC_CHANGE, c);
		goto out;
	}

	/*
	 * periodically send out the location
	 *
	 * Even if:
	 *  - the grid didn't change
	 *  - we don't have a fix
	 *
	 * This way, we're not silent for long periods without changes to
	 * the fix.
	 */
	notify(ANNOUNCE_LOC_CURRENT, c);

out:
	*p = *c;
}

static bool __gps_connect(void)
{
	/*
	 * To avoid spamming the log with reconnection attempts, we only
	 * print the error messages only once until we get a connection.
	 * Then, if we lose the connection, we print the messages once.  And
	 * so on.
	 */
	static bool silence;

	if (connected)
		return true;

	if (gps_open(HOSTNAME, PORT, &gps)) {
		if (!silence)
			cmn_err(CE_WARN, "Connection to gpsd failed: %s",
				gps_errstr(errno));
		silence = true;
		return false;
	}

	if (gps_stream(&gps, WATCH_ENABLE | WATCH_JSON, NULL)) {
		cmn_err(CE_WARN, "Failed to enable gpsd watches: %s",
			gps_errstr(errno));

		if (gps_close(&gps))
			cmn_err(CE_ERROR, "Failed to close gpsd connection: %s",
				gps_errstr(errno));

		silence = true;
		return false;
	}

	cmn_err(CE_INFO, "Connected, acquiring location info");

	connected = true;
	silence = false;

	/* reset previous fix data */
	memset(&prevfix, 0, sizeof(prevfix));

	return true;
}

/*
 * returns:
 *   -1 on error
 *    0 if no data available
 *    1 if data availabl
 */
static int __gps_read(void)
{
	int ret;

	if (!gps_waiting(&gps, 500))
		return 0;

	errno = 0;
	ret = gps_read(&gps, NULL, 0);
	if (ret < 0) {
		connected = false;
		notify(ANNOUNCE_LOC_DISCONNECT, NULL);
		return -1; /* error */
	} else if (!ret) {
		return 0; /* no data available */
	} else {
		return 1; /* data available */
	}
}

static void *location_watcher(void *unused)
{
	for (;;) {
		/* connect as necessary */
		if (!__gps_connect()) {
			/* failed - sleep for a while and try again */
			sleep(RECONNECT_DELAY);
			continue;
		}

		switch (__gps_read()) {
			case -1:
			case 0:
				sleep(1); /* nothing to do */
				break;
			case 1:
				process();
				break;
		}
	}

	cmn_err(CE_INFO, "Thread exiting...");

	notify(ANNOUNCE_LOC_DISCONNECT, NULL);

	/* clean up libgps */
	if (gps_close(&gps))
		cmn_err(CE_ERROR, "Failed to close gpsd connection: %s",
			strerror(errno));

	return NULL;
}

int location_start(void)
{
	int ret;

	ret = xthr_create_named(&location_thread, "location",
				location_watcher, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to start location thread: %s",
			xstrerror(ret));

	return ret;
}

void location_stop(void)
{
	int ret;

	ret = xthr_join(location_thread, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to join location thread: %s",
			xstrerror(ret));
}
